package prtg

import (
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"log"
	"net"
	"net/http"
	"net/url"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/it-novum/event-collectd/helper"
	"github.com/it-novum/event-collectd/oitc"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

type PrtgHttpClient struct {
	ctx                   context.Context
	config                Config
	client                http.Client
	checkResultChannel    chan *oitc.CheckResult
	rawCommandChannel     chan *oitc.BrokerRawCommand
	cancelDowntimeChannel chan *oitc.CancelDowntime
	mapping               *oitc.OpenITCOCKPITMapping
	wg                    sync.WaitGroup
	shutdown              chan struct{}
}

var LastSensorCheckresults map[int64]*Sensor

// New creates a HTTP client to query the PRTG HTTP API v1
// https://www.paessler.com/manuals/prtg/http_api
// See api_examples.md for plain curl examples
func New(config Config) (*PrtgHttpClient, error) {
	var proxyFunc func(*http.Request) (*url.URL, error) = nil
	if config.ProxyAddress != "" {
		proxyURL, err := url.Parse(config.ProxyAddress)
		if err != nil {
			return nil, err
		}
		proxyFunc = http.ProxyURL(proxyURL)
	}

	return &PrtgHttpClient{
		config: config,
		client: http.Client{
			Timeout: 0, //5 * time.Minute,
			Transport: &http.Transport{
				TLSHandshakeTimeout: 10 * time.Second,
				TLSClientConfig: &tls.Config{
					InsecureSkipVerify: config.IgnoreSslCertificate == 1,
				},
				Proxy: proxyFunc,
				DialContext: (&net.Dialer{
					Timeout:   30 * time.Second,
					KeepAlive: 30 * time.Second,
				}).DialContext,
			},
		},
	}, nil
}

func (c *PrtgHttpClient) Close() {
	close(c.shutdown)
	c.wg.Wait()
}

func (c *PrtgHttpClient) startApiPoller() error {
	defer c.wg.Done()

	// Golangs ticker does not fire immediately so we need a "fake" tick.
	ticker := time.NewTicker(time.Duration(c.config.PollingInterval) * time.Second)
	firstTick := make(chan struct{}, 1)
	firstTick <- struct{}{}

	defer ticker.Stop()
	defer close(firstTick)

	var err error
	for {

		select {

		case <-firstTick:
			// Fire the poller immediately
			log.Println("[PRTG] Initial poll of PRTG API")
			err = c.PollPrtgApi()

			if err == context.Canceled {
				log.Println("[PRTG] Context canceled. Shutdown PRTG Poller")
				return nil
			}

			if err != nil {
				log.Println("[PRTG] PRTG Poller error:", err)
				log.Print("[PRTG] Trying to reconnect on next polling interval tick")
			}

		case <-ticker.C:
			// New Tick, poll the PRTG API
			if c.config.Verbose {
				log.Println("[PRTG] New Tick - poll the PRTG API")
			}
			err = c.PollPrtgApi()

			if err == context.Canceled {
				log.Println("[PRTG] Context canceled. Shutdown PRTG Poller")
				return nil
			}

			if err != nil {
				log.Println("[PRTG] PRTG Poller error:", err)
				log.Print("[PRTG] Trying to reconnect on next polling interval tick")
			}

		case <-c.ctx.Done():
			// Close PRTG Poller
			return nil
		}
	}
}

func (c *PrtgHttpClient) PollPrtgApi() error {

	if c.config.Verbose {
		log.Println("[PRTG] Connecting to PRTG")
	}

	deviceCount, err := c.getDeviceCount()
	if err != nil {
		//log.Println(err)
		return err
	}

	sensorCount, err := c.getSensorCount()
	if err != nil {
		//log.Println(err)
		return err
	}

	// initialize hash map
	if LastSensorCheckresults == nil {
		LastSensorCheckresults = make(map[int64]*Sensor)
	}

	if err != nil {
		log.Println("[PRTG] Trying to reconnect to PRTG API in 5 seconds")
		time.Sleep(5 * time.Second) // Do not DDOS the PRTG API
	}

	if err == nil && sensorCount > 0 {
		// PRTG API connection successfully
		devices, err := c.getAllDevices(deviceCount)
		if err != nil {
			log.Println("[PRTG] Error while querying device status")
			return err
		}

		// Set device status as host status in openITCOCKPIT
		for _, device := range devices {
			hostUuid, err := c.mapping.GetHostUuidByHostIdentifier(strconv.FormatInt(device.ObjectId, 10))
			if err != nil {
				if c.config.Verbose {
					log.Println("[PRTG] ", err)
				}
				continue
			}

			// Send host check result to FIFO
			c.checkResultChannel <- &oitc.CheckResult{
				HostName:           hostUuid,
				ServiceDescription: "",
				Output:             device.Message,
				LongOutput:         "", // Not available in PRTG
				PerfData:           "", // Not available in PRTG
				CheckType:          oitc.HOST_CHECK,
				ReturnCode:         c.mapPrtgDeviceStateToOitcHostState(device.StatusId),
				StartTime:          time.Now().Unix(),
				EndTime:            time.Now().Unix(),
				EarlyTimeout:       0,
				Latency:            0.0,
				ExitedOk:           1,
			}

		}

		sensors, err := c.getAllSensors(sensorCount)
		if err != nil {
			log.Println("[PRTG] Error while querying sensor status")
			return err
		}

		for _, sensor := range sensors {
			if lastSensorValue, ok := LastSensorCheckresults[sensor.ObjectId]; ok {
				// Make sure that the sensor got checked since we processed the last result
				if sensor.Lastcheck <= lastSensorValue.Lastcheck {
					// Timestamp is less or eq the last timestamp we processed, no new check result
					// todo: maybe check for ack or downtimes?
					continue
				}
			}

			LastSensorCheckresults[sensor.ObjectId] = sensor

			var channels ChannelsApiResult
			if c.config.IncludeChannels == 1 {
				// The PRTG API is not very good. We need to query the channels for each sensor separately.
				// Probably this is the moment where PRTG is stating to melt
				// Denn so zerbröselt der Keks nun mal
				channels, err = c.getChannelsBySensorId(sensor.ObjectId)
				if err != nil {
					log.Fatalln("[PRTG] Error while loading channels:", err)
				}
			}

			// todo Implement Lookup by Sensor Object Id
			// Convert int64 id into a String
			deviceId := strconv.FormatInt(sensor.DeviceId, 10)
			sensorId := strconv.FormatInt(sensor.ObjectId, 10)

			hostUuid, err := c.mapping.GetHostUuidByHostIdentifier(deviceId)
			if err != nil {
				if c.config.Verbose {
					log.Println("[PRTG] ", err)
				}
				continue
			}

			serviceUuid, err := c.mapping.GetServiceUuidByHostIdentifierAndServiceIdentifier(deviceId, sensorId)
			if err != nil {
				if c.config.Verbose {
					log.Println("[PRTG] ", err)
				}
				continue
			}

			// Send service check result to FIFO
			c.checkResultChannel <- &oitc.CheckResult{
				HostName:           hostUuid,
				ServiceDescription: serviceUuid,
				Output:             sensor.Message,
				LongOutput:         getLongoutputForChannels(channels.Channels),
				PerfData:           getPerformanceDataAsString(channels.Channels),
				CheckType:          oitc.SERVICE_CHECK,
				ReturnCode:         c.mapPrtgSensorStateToOitcServiceState(sensor.StatusId),
				StartTime:          helper.ExcelDateToUnix(sensor.Lastcheck),
				EndTime:            helper.ExcelDateToUnix(sensor.Lastcheck),
				EarlyTimeout:       0,
				Latency:            0.0,
				ExitedOk:           1,
			}

		}
	}

	return nil
}

func (c *PrtgHttpClient) getDeviceCount() (int64, error) {
	baseUrl := c.config.ApiUrl + "/api/table.json"

	// Create a URL that looks like this
	// /api/table.json?content=devices&count=0&apitoken=BV3PKNYJHAUEQIR7UTCZWFYUN7FKNTO62DFBUFQZ5M======

	u, err := url.Parse(baseUrl)
	if err != nil {
		return 0, err
	}

	queryParams := u.Query()
	queryParams.Set("content", "devices")
	queryParams.Set("count", "0")
	if c.config.ApiToken != "" {
		queryParams.Set("apitoken", c.config.ApiToken)
	} else {
		queryParams.Set("username", c.config.ApiUser)
		queryParams.Set("password", c.config.ApiPassword)
	}
	u.RawQuery = queryParams.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return 0, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[PRTG] Server error: ", err)
		return 0, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[PRTG] Error - Non-OK status code:", response.Status)
		return 0, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[PRTG] Error reading response body:", err)
		return 0, err
	}

	var result VersionAndObjectCount
	if err := json.Unmarshal(body, &result); err != nil {
		log.Println("[PRTG] Error wile parsing response json:", err)
		return 0, err
	}

	return result.TreeSize, nil
}

func (c *PrtgHttpClient) getSensorCount() (int64, error) {
	baseUrl := c.config.ApiUrl + "/api/table.json"

	// Create a URL that looks like this
	// /api/table.json?content=sensors&count=0&apitoken=BV3PKNYJHAUEQIR7UTCZWFYUN7FKNTO62DFBUFQZ5M======

	u, err := url.Parse(baseUrl)
	if err != nil {
		return 0, err
	}

	queryParams := u.Query()
	queryParams.Set("content", "sensors")
	queryParams.Set("count", "0")
	if c.config.ApiToken != "" {
		queryParams.Set("apitoken", c.config.ApiToken)
	} else {
		queryParams.Set("username", c.config.ApiUser)
		queryParams.Set("password", c.config.ApiPassword)
	}
	u.RawQuery = queryParams.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return 0, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[PRTG] Server error: ", err)
		return 0, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[PRTG] Error - Non-OK status code:", response.Status)
		return 0, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[PRTG] Error reading response body:", err)
		return 0, err
	}

	var result VersionAndObjectCount
	if err := json.Unmarshal(body, &result); err != nil {
		log.Println("[PRTG] Error wile parsing response json:", err)
		return 0, err
	}

	return result.TreeSize, nil
}

func (c *PrtgHttpClient) getAllDevices(deviceCount int64) (map[int64]*Device, error) {
	baseUrl := c.config.ApiUrl + "/api/table.json"

	// Create a URL that looks like this
	// /api/table.json?content=devices&columns=objid,name,active,,status,message&apitoken=BV3PKNYJHAUEQIR7UTCZWFYUN7FKNTO62DFBUFQZ5M======

	u, err := url.Parse(baseUrl)
	if err != nil {
		return nil, err
	}

	queryParams := u.Query()
	queryParams.Set("content", "devices")
	queryParams.Set("columns", "objid,name,active,,status,message")
	queryParams.Set("count", fmt.Sprint(deviceCount))
	if c.config.ApiToken != "" {
		queryParams.Set("apitoken", c.config.ApiToken)
	} else {
		queryParams.Set("username", c.config.ApiUser)
		queryParams.Set("password", c.config.ApiPassword)
	}
	u.RawQuery = queryParams.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return nil, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[PRTG] Server error: ", err)
		return nil, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[PRTG] Error - Non-OK status code:", response.Status)
		return nil, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[PRTG] Error reading response body:", err)
		return nil, err
	}

	var result DevicesApiResult
	if err := json.Unmarshal(body, &result); err != nil {
		log.Println("[PRTG] Error wile parsing response json:", err)
		return nil, err
	}

	// Store all devices as hashmap for easy access in sensors loop
	devices := make(map[int64]*Device, len(result.Devices))
	for _, d := range result.Devices {
		devices[d.ObjectId] = d
	}

	return devices, nil
}

func (c *PrtgHttpClient) getAllSensors(sensorCount int64) ([]*Sensor, error) {
	baseUrl := c.config.ApiUrl + "/api/table.json"

	// Create a URL that looks like this
	// /api/table.json?content=sensors&columns=objid,type,name,active,downtime,downtimetime,sensor,lastcheck,device,probe,status,message,lastup&apitoken=BV3PKNYJHAUEQIR7UTCZWFYUN7FKNTO62DFBUFQZ5M======

	u, err := url.Parse(baseUrl)
	if err != nil {
		return nil, err
	}

	queryParams := u.Query()
	queryParams.Set("content", "sensors")
	queryParams.Set("columns", "objid,type,name,active,downtime,downtimetime,sensor,lastcheck,device,probe,status,message,parentid")
	queryParams.Set("count", fmt.Sprint(sensorCount))
	if c.config.ApiToken != "" {
		queryParams.Set("apitoken", c.config.ApiToken)
	} else {
		queryParams.Set("username", c.config.ApiUser)
		queryParams.Set("password", c.config.ApiPassword)
	}
	u.RawQuery = queryParams.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return nil, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[PRTG] Server error: ", err)
		return nil, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[PRTG] Error - Non-OK status code:", response.Status)
		return nil, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[PRTG] Error reading response body:", err)
		return nil, err
	}

	var result SensorsApiResult
	if err := json.Unmarshal(body, &result); err != nil {
		log.Println("[PRTG] Error wile parsing response json:", err)
		return nil, err
	}

	return result.Sensors, nil
}

func (c *PrtgHttpClient) getChannelsBySensorId(sensorId int64) (ChannelsApiResult, error) {
	baseUrl := c.config.ApiUrl + "/api/table.json"

	// Create a URL that looks like this
	// /api/table.json?content=sensors&columns=objid,type,name,active,downtime,downtimetime,sensor,lastcheck,device,probe,status,message,lastup&apitoken=BV3PKNYJHAUEQIR7UTCZWFYUN7FKNTO62DFBUFQZ5M======

	u, err := url.Parse(baseUrl)
	if err != nil {
		return ChannelsApiResult{}, err
	}

	queryParams := u.Query()
	queryParams.Set("content", "channels")
	queryParams.Set("columns", "objid,name=textraw,info=treejson,minimum,maximum,condition,lastvalue")
	queryParams.Set("id", fmt.Sprint(sensorId))
	if c.config.ApiToken != "" {
		queryParams.Set("apitoken", c.config.ApiToken)
	} else {
		queryParams.Set("username", c.config.ApiUser)
		queryParams.Set("password", c.config.ApiPassword)
	}
	u.RawQuery = queryParams.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return ChannelsApiResult{}, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[PRTG] Server error: ", err)
		return ChannelsApiResult{}, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[PRTG] Error - Non-OK status code:", response.Status)
		return ChannelsApiResult{}, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[PRTG] Error reading response body:", err)
		return ChannelsApiResult{}, err
	}

	var result ChannelsApiResult
	if err := json.Unmarshal(body, &result); err != nil {
		log.Println("[PRTG] Error wile parsing response json:", err)
		return ChannelsApiResult{}, err
	}

	return result, nil
}

func (c *PrtgHttpClient) mapPrtgDeviceStateToOitcHostState(prtgState int64) int64 {
	switch prtgState {

	case PRTG_SCANNING,
		PRTG_UP,
		PRTG_PAUSED_BY_USER,
		PRTG_PAUSED_BY_DEPENDENCY,
		PRTG_PAUSED_BY_SCHEDULE,
		PRTG_PAUSED_UNTIL,
		PRTG_DOWN_PARTIAL:
		return oitc.STATE_UP

	case PRTG_WARNING,
		PRTG_UNUSUAL,
		PRTG_DOWN,
		PRTG_DOWN_ACKNOWLEDGED:
		return oitc.STATE_DOWN

	case PRTG_NONE,
		PRTG_UNKNOWN,
		PRTG_NO_PROBE,
		PRTG_NOT_LICENSE:
		return oitc.STATE_UNREACHABLE
	}

	return oitc.STATE_UNREACHABLE
}

func (c *PrtgHttpClient) mapPrtgSensorStateToOitcServiceState(prtgState int64) int64 {
	switch prtgState {

	case PRTG_SCANNING,
		PRTG_UP,
		PRTG_PAUSED_BY_USER,
		PRTG_PAUSED_BY_DEPENDENCY,
		PRTG_PAUSED_BY_SCHEDULE,
		PRTG_PAUSED_UNTIL,
		PRTG_DOWN_PARTIAL:
		return oitc.STATE_OK

	case PRTG_WARNING,
		PRTG_UNUSUAL:
		return oitc.STATE_WARNING

	case PRTG_DOWN,
		PRTG_DOWN_ACKNOWLEDGED:
		return oitc.STATE_CRITICAL

	case PRTG_NONE,
		PRTG_UNKNOWN,
		PRTG_NO_PROBE,
		PRTG_NOT_LICENSE:
		return oitc.STATE_UNKNOWN
	}

	return oitc.STATE_UNKNOWN
}

func getPerformanceDataAsString(channels []*Channel) string {
	var prtgPerfdataStringArray []string

	if len(channels) > 0 {
		// We need to re-write the performance data into something Naemon (and Nagios) can understand.

		prtgPerfdataStringArray = make([]string, len(channels))
		for i, channel := range channels {

			isPercent := false
			if len(channel.Info.Data) >= 1 {
				isPercent = channel.Info.Data[0].IsPercent

				if channel.Info.Data[0].ObjectId < 0 {
					// Skip PRTG internal channels such as PRTG_CHANNEL_DOWNTIME (-4) etc
					continue
				}
			}

			var perfdataString string
			if isPercent {
				perfdataString = fmt.Sprintf(
					"'%v'=%v%v;%v;%v;%v;%v",
					channel.Name,
					channel.LastvalueRaw,
					"%",
					"",    // Warning
					"",    // Critical
					"0",   // Min
					"100", // Max
				)
			} else {
				// Unknown unit (The PRTG APIv1 has no method to query raw value + unit)
				perfdataString = fmt.Sprintf(
					"'%v'=%v%v;%v;%v;%v;%v",
					channel.Name,
					channel.LastvalueRaw,
					"",
					"", // Warning
					"", // Critical
					"", // Min
					"", // Max
				)
			}

			prtgPerfdataStringArray[i] = perfdataString
		}
	}

	if len(prtgPerfdataStringArray) > 0 {
		return strings.Join(prtgPerfdataStringArray[:], " ")
	}

	// No perfdata
	return ""
}

func getLongoutputForChannels(channels []*Channel) string {
	var longOutput []string

	if len(channels) > 0 {
		// We need to re-write the performance data into something Naemon (and Nagios) can understand.

		longOutput = make([]string, 0, len(channels))
		for _, channel := range channels {

			var lastvalue string = channel.LastvalueRaw.String()
			if len(channel.Info.Data) >= 1 {
				if channel.Info.Data[0].ObjectId < 0 {
					// Skip PRTG internal channels such as PRTG_CHANNEL_DOWNTIME (-4) etc
					continue
				}

				// Use the formated value such as "0,03 Mbit/s"
				if channel.Info.Data[0].Lastvalue != nil {
					lastvalue = channel.Info.Data[0].Lastvalue.ToString()
				}
			}

			line := fmt.Sprintf(
				"%v: %v",
				channel.Name,
				lastvalue,
			)

			longOutput = append(longOutput, line)
		}
	}

	if len(longOutput) > 0 {
		return strings.Join(longOutput[:], "\n")
	}

	// No long output
	return ""
}

// Start reading from PRTG API returns immediately (SHOULD NOT RUN IN GOROUTINE)
// Due to there is no stream we can subscribe, we have to poll the API regularly
func (c *PrtgHttpClient) Start(ctx context.Context, cr chan *oitc.CheckResult, raw chan *oitc.BrokerRawCommand, cancelDowntimeChannel chan *oitc.CancelDowntime, mapping *oitc.OpenITCOCKPITMapping) error {
	c.shutdown = make(chan struct{})
	c.checkResultChannel = cr
	c.rawCommandChannel = raw
	c.cancelDowntimeChannel = cancelDowntimeChannel
	c.mapping = mapping

	c.wg.Add(1)
	go func() {
		ctx, cancel := context.WithCancel(ctx)
		c.ctx = ctx

		defer func() {
			cancel()
			c.wg.Done()
		}()

		// Start a new thread (go func) that will poll the PRTG API regularly
		c.wg.Add(1)
		go c.startApiPoller()

		for {
			select {
			case <-ctx.Done():
				return

			case _, ok := <-c.shutdown:
				if !ok {
					// Channel c.shutdown got closed - close this PRTG poller
					log.Println("[PRTG] Shutdown PRTG poller")
					return
				}
			}

		}

	}()

	return nil
}
